@@ -1,6 +1,6 @@
 package SVN::Mirror::Ra;
 @ISA = ('SVN::Mirror');
-$VERSION = '0.35';
+$VERSION = '0.66';
 use strict;
 use SVN::Core;
 use SVN::Repos;
@@ -8,6 +8,7 @@ use SVN::Fs;
 use SVN::Delta;
 use SVN::Ra;
 use SVN::Client ();
+use constant OK => $SVN::_Core::SVN_NO_ERROR;
 
 sub new {
     my $class = shift;
@@ -15,120 +16,373 @@ sub new {
     %$self = @_;
     $self->{source} =~ s{/+$}{}g;
 
-    my ($root, $path) = split ('!', $self->{source});
-
     @{$self}{qw/source source_root source_path/} =
-	(join('', $root, $path), $root, $path)
-	    if defined $path;
+	_parse_source ($self->{source});
+
+    @{$self}{qw/rsource rsource_root rsource_path/} =
+	_parse_source ($self->{rsource}) if $self->{rsource};
 
     return $self;
 }
 
+sub _parse_source {
+    my $source = shift;
+    my ($root, $path) = split ('!', $source, 2);
+    $path ||= '';
+    return (join('', $root, $path), $root, $path)
+}
+
+sub _store_source {
+    my ($root, $path) = @_;
+    return join('!', $root, $path);
+}
+
+sub _get_prop {
+    my ($self, $ra, $path, $propname) = @_;
+}
+
+sub _is_descendent {
+    my ($parent, $child) = @_;
+    return 1 if $parent eq $child;
+    $parent = "$parent/" unless $parent eq '/';
+    return $parent eq substr ($child, 0, length ($parent));
+}
+
+sub _check_overlap {
+    my ($self) = @_;
+    my $fs = $self->{repos}->fs;
+    my $root = $fs->revision_root ($fs->youngest_rev);
+    for (map {$root->node_prop ($_, 'svm:source')} SVN::Mirror::list_mirror ($self->{repos})) {
+	my (undef, $source_root, $source_path) = _parse_source ($_);
+	next if $source_root ne $self->{source_root};
+	die "Mirroring overlapping paths not supported\n"
+	    if _is_descendent ($source_path, $self->{source_path})
+	    || _is_descendent ($self->{source_path}, $source_path);
+    }
+}
+
 sub init_state {
     my ($self, $txn) = @_;
-    my $ra = $self->_new_ra;
+    my $ra = $self->_new_ra (url => $self->{source});
 
     my $uuid = $self->{source_uuid} = $ra->get_uuid ();
     my $source_root = $ra->get_repos_root ();
-    my $txnroot = $txn->root;
-
-    $txnroot->change_node_prop ($self->{target_path},
-				'svm:uuid', "$uuid");
-
     my $path = $self->{source};
-    die "source url not under source root"
+    $txn->abort, die "source url not under source root"
 	if substr($path, 0, length($source_root), '') ne $source_root;
 
     $self->{source_root} = $source_root;
     $self->{source_path} = $path;
     $self->{fromrev} = 0;
 
-    return join('!', $source_root, $path);
+    # XXX: abort txn before dying
+    $self->_check_overlap;
+
+    # check if the url exists
+    if ($ra->check_path ('', -1) != $SVN::Node::dir) {
+	$txn->abort;
+	die "$self->{source} is not a directory.\n";
+    }
+    unless ($self->{source} eq $self->{source_root}) {
+	undef $ra; # bizzare perlgc
+	$ra = $self->_new_ra (url => $self->{source_root});
+    }
+
+    # check if mirror source is already a mirror
+    # older SVN::RA will return Reporter so prop would be undef
+    my (undef, undef, $prop) = $ra->get_dir ('', -1);
+    if ($prop && $prop->{'svm:mirror'}) {
+	my $rroot;
+	for ($prop->{'svm:mirror'} =~ m/^.*$/mg) {
+	    if (_is_descendent ($_, $self->{source_path})) {
+		$rroot = $_;
+		last;
+	    }
+	    elsif (_is_descendent ($self->{source_path}, $_)) {
+		$txn->abort, die "Can't relay mirror outside mirror anchor $_";
+	    }
+	}
+	if ($rroot) {
+	    $rroot =~ s|^/||;
+	    (undef, undef, $prop) = $ra->get_dir ($rroot, -1);
+	    $txn->abort, die "relayed mirror source doesn't not have svm:source"
+		unless exists $prop->{'svm:source'};
+	    @{$self}{qw/rsource rsource_root rsource_path/} =
+		@{$self}{qw/source source_root source_path/};
+	    $self->{rsource_uuid} = $uuid;
+	    $self->{source_path} =~ s|^/\Q$rroot\E||;
+	    @{$self}{qw/source source_uuid/} = @{$prop}{qw/svm:source svm:uuid/};
+	    $self->{source} .= '!' if index ($self->{source}, '!') == -1;
+	    @{$self}{qw/source source_root source_path/} =
+		_parse_source ($self->{source}.$self->{source_path});
+
+	    $txn->abort, die "relayed source and source have same repository uuid"
+		if $self->{source_uuid} eq $self->{rsource_uuid};
+
+	    my $txnroot = $txn->root;
+	    $txnroot->change_node_prop ($self->{target_path}, 'svm:rsource',
+					_store_source ($source_root, $path));
+	    $txnroot->change_node_prop ($self->{target_path}, 'svm:ruuid',
+					$uuid);
+	    $txn->change_prop ("svm:headrev", "$self->{rsource_uuid}:$self->{fromrev}\n");
+
+	    return _store_source ($self->{source_root}, $self->{source_path});
+	}
+    }
+
+    @{$self}{qw/rsource rsource_root rsource_path/} =
+	@{$self}{qw/source source_root source_path/};
+
+    $self->{rsource_uuid} = $self->{source_uuid};
+
+    $txn->change_prop ("svm:headrev", "$self->{rsource_uuid}:$self->{fromrev}\n");
+    return _store_source ($source_root, $path);
 }
 
 sub load_state {
-    my ($self, $txn) = @_;
+    my ($self) = @_;
 
-    my $txnroot = $txn->root;
-    my $changed = $self->{root}->node_created_rev ($self->{target_path});
-
-    my $prop = $self->{fs}->revision_proplist ($changed);
-    die "no headrev for $self->{source} at rev $changed"
-	unless exists $prop->{"svm:headrev:$self->{source}"};
-    $self->{fromrev} = $prop->{"svm:headrev:$self->{source}"};
-    $self->{VSN} = $prop->{"svm:vsnroot:$self->{source}"};
-    # upgrade for 0.27
-    $self->{source_uuid} = $txnroot->node_prop ($self->{target_path}, 'svm:uuid');
-    my $info = "svm:mirror:$self->{source_uuid}:".($self->{source_path} || '/');
-
-    unless ($txnroot->node_prop ('/', $info)) {
-	$txnroot->change_node_prop ('/', $info, $self->{target_path});
-	my (undef, $rev) = $txn->commit ();
-	$self->{fs}->change_rev_prop ($rev, "svm:headrev:$self->{source}", $self->{fromrev});
-	$self->{fs}->change_rev_prop ($rev, "svn:author", 'svm');
-	$self->{fs}->change_rev_prop
-		    ($rev, "svn:log", "SVM: upgraded info for $self->{target_path}.");
+    my $prop = $self->{root}->node_proplist ($self->{target_path});
+    @{$self}{qw/source_uuid rsource_uuid/} =
+	@{$prop}{qw/svm:uuid svm:ruuid/};
+    unless ($self->{rsource}) {
+	@{$self}{qw/rsource rsource_root rsource_path/} =
+	    @{$self}{qw/source source_root source_path/};
+	$self->{rsource_uuid} = $self->{source_uuid};
     }
-    else {
-	$txn->abort ();
+
+    die "please upgrade the mirror state\n"
+	if $self->{root}->node_prop ('/', join (':', 'svm:mirror', $self->{source_uuid},
+						$self->{source_path} || '/'));
+
+    unless ($self->{ignore_lock}) {
+	die "no headrev"
+	    unless defined $self->load_fromrev;
     }
+    return;
 }
 
 sub _new_ra {
     my ($self, %arg) = @_;
-    SVN::Ra->new( url => $self->{source},
+    $self->{config} ||= SVN::Core::config_get_config (undef, $self->{pool});
+    $self->{auth} ||= $self->_new_auth;
+    SVN::Ra->new( url => $self->{rsource},
 		  auth => $self->{auth},
-		  pool => $self->{pool},
 		  config => $self->{config},
 		  %arg);
 }
 
+sub _new_auth {
+    my ($self) = @_;
+    # create a subpool that is not automatically destroyed
+    my $pool = SVN::Pool::create (${$self->{pool}});
+    $pool->default;
+    my ($baton, $ref) = SVN::Core::auth_open_helper([
+        SVN::Client::get_simple_provider (),
+        SVN::Client::get_ssl_server_trust_file_provider (),
+        SVN::Client::get_username_provider (),
+        SVN::Client::get_simple_prompt_provider( $self->can('_simple_prompt'), 2),
+        SVN::Client::get_ssl_server_trust_prompt_provider( $self->can('_ssl_server_trust_prompt') ),
+        SVN::Client::get_ssl_client_cert_prompt_provider( $self->can('_ssl_client_cert_prompt'), 2 ),
+        SVN::Client::get_ssl_client_cert_pw_prompt_provider( $self->can('_ssl_client_cert_pw_prompt'), 2 ),
+        SVN::Client::get_username_prompt_provider( $self->can('_username_prompt'), 2),
+    ]);
+    $self->{auth_ref} = $ref;
+    return $baton;
+}
+
+sub _simple_prompt {
+    my ($cred, $realm, $default_username, $may_save, $pool) = @_;
+
+    if (defined $default_username and length $default_username) {
+        print "Authentication realm: $realm\n" if defined $realm and length $realm;
+        $cred->username($default_username);
+    }
+    else {
+        _username_prompt($cred, $realm, $may_save, $pool);
+    }
+
+    $cred->password(_read_password("Password for '" . $cred->username . "': "));
+    $cred->may_save($may_save);
+
+    return OK;
+}
+
+sub _ssl_server_trust_prompt {
+    my ($cred, $realm, $failures, $cert_info, $may_save, $pool) = @_;
+
+    print "Error validating server certificate for '$realm':\n";
+
+    print " - The certificate is not issued by a trusted authority. Use the\n",
+          "   fingerprint to validate the certificate manually!\n"
+      if ($failures & $SVN::Auth::SSL::UNKNOWNCA);
+
+    print " - The certificate hostname does not match.\n"
+      if ($failures & $SVN::Auth::SSL::CNMISMATCH);
+
+    print " - The certificate is not yet valid.\n"
+      if ($failures & $SVN::Auth::SSL::NOTYETVALID);
+
+    print " - The certificate has expired.\n"
+      if ($failures & $SVN::Auth::SSL::EXPIRED);
+
+    print " - The certificate has an unknown error.\n"
+      if ($failures & $SVN::Auth::SSL::OTHER);
+
+    printf(
+        "Certificate information:\n".
+        " - Hostname: %s\n".
+        " - Valid: from %s until %s\n".
+        " - Issuer: %s\n".
+        " - Fingerprint: %s\n",
+        map $cert_info->$_, qw(hostname valid_from valid_until issuer_dname fingerprint)
+    );
+
+    print(
+        $may_save
+            ? "(R)eject, accept (t)emporarily or accept (p)ermanently? "
+            : "(R)eject or accept (t)emporarily? "
+    );
+
+    my $choice = lc(substr(<STDIN> || 'R', 0, 1));
+
+    if ($choice eq 't') {
+        $cred->may_save(0);
+        $cred->accepted_failures($failures);
+    }
+    elsif ($may_save and $choice eq 'p') {
+        $cred->may_save(1);
+        $cred->accepted_failures($failures);
+    }
+
+    return OK;
+}
+
+sub _ssl_client_cert_prompt {
+    my ($cred, $realm, $may_save, $pool) = @_;
+
+    print "Client certificate filename: ";
+    chomp(my $filename = <STDIN>);
+    $cred->cert_file($filename);
+
+    return OK;
+}
+
+sub _ssl_client_cert_pw_prompt {
+    my ($cred, $realm, $may_save, $pool) = @_;
+
+    $cred->password(_read_password("Passphrase for '%s': "));
+
+    return OK;
+}
+
+sub _username_prompt {
+    my ($cred, $realm, $may_save, $pool) = @_;
+
+    print "Authentication realm: $realm\n" if defined $realm and length $realm;
+    print "Username: ";
+    chomp(my $username = <STDIN>);
+    $username = '' unless defined $username;
+
+    $cred->username($username);
+
+    return OK;
+}
+
+sub _read_password {
+    my ($prompt) = @_;
+
+    print $prompt;
+
+    require Term::ReadKey;
+    Term::ReadKey::ReadMode('noecho');
+
+    my $password = '';
+    while (defined(my $key = Term::ReadKey::ReadKey(0))) {
+        last if $key =~ /[\012\015]/;
+        $password .= $key;
+    }
+
+    Term::ReadKey::ReadMode('restore');
+    print "\n";
+
+    return $password;
+}
+
+sub _revmap {
+    my ($self, $rev, $ra) = @_;
+    $ra ||= $self->{cached_ra};
+    $SVN::Core::VERSION ge '1.1.0' ?
+	$ra->rev_prop ($rev, 'svm:headrev') :
+	$ra->rev_proplist ($rev)->{'svm:headrev'};
+}
+
 sub committed {
-    my ($self, $date, $sourcerev, $rev, undef, undef, $pool) = @_;
-    my $cpool = SVN::Pool->new_default ($pool);
+    my ($self, $revmap, $date, $sourcerev, $rev) = @_;
     $self->{fs}->change_rev_prop($rev, 'svn:date', $date);
-    $self->{fs}->change_rev_prop($rev, "svm:headrev:$self->{source}",
-				 "$sourcerev",);
-    $self->{fs}->change_rev_prop($rev, "svm:vsnroot:$self->{source}",
-				 "$self->{VSN}") if $self->{VSN};
+    # sync remote headrev too
+    $self->{fs}->change_rev_prop($rev, 'svm:headrev', $revmap."$self->{rsource_uuid}:$sourcerev\n");
+    $self->{fs}->change_rev_prop($rev, 'svm:incomplete', '*')
+	if $self->{rev_incomplete};
     $self->{headrev} = $rev;
 
+    $self->unlock ('mirror');
     print "Committed revision $rev from revision $sourcerev.\n";
 }
 
+our $debug;
+
 sub mirror {
     my ($self, $fromrev, $paths, $rev, $author, $date, $msg, $ppool) = @_;
     my $ra;
 
+    if ($debug and eval { require BSD::Resource; 1 }) {
+	my ($usertime, $systemtime,
+	    $maxrss, $ixrss, $idrss, $isrss, $minflt, $majflt, $nswap,
+	    $inblock, $oublock, $msgsnd, $msgrcv,
+	    $nsignals, $nvcsw, $nivcsw) = BSD::Resource::getrusage();
+	print ">>> mirroring $rev:\n";
+	print ">>> $usertime $systemtime $maxrss $ixrss $idrss $isrss\n";
+    }
+
     my $pool = SVN::Pool->new_default ($ppool);
-    my $newrev;
-    use SVN::Mirror::Ra;
+    my ($newrev, $revmap);
+
+    $ra = $self->{cached_ra}
+	if exists $self->{cached_ra_url} &&
+	    $self->{cached_ra_url} eq $self->{rsource};
+    if ($ra && $self->{rsource} =~ m/^http/ && --$self->{cached_life} == 0) {
+	undef $ra;
+    }
+    $ra ||= $self->_new_ra;
+
+    $revmap = $self->_revmap ($rev, $ra) if $self->_relayed;
+    $revmap ||= '';
+
     my $editor = SVN::Mirror::Ra::MirrorEditor->new
 	($self->{repos}->get_commit_editor
 	 ('', $self->{target_path}, $author, $msg,
-	  sub { $newrev = $_[0]; $self->committed($date, $rev, @_) }));
+	  sub { $newrev = $_[0];
+		$self->committed ($revmap, $date, $rev, @_) }));
 
     $self->{working} = $rev;
     $editor->{mirror} = $self;
 
-    $ra = $self->{cached_ra}
-	if exists $self->{cached_ra_url} &&
-	    $self->{cached_ra_url} eq $self->{source};
-
-    $ra ||= $self->_new_ra ( pool => SVN::Pool->new );
-    @{$self}{qw/cached_ra cached_ra_url/} = ($ra, $self->{source});
-
+    @{$self}{qw/cached_ra cached_ra_url/} = ($ra, $self->{rsource});
     if ( ( $fromrev == 0
-           || !(defined $fromrev && $self->find_local_rev($fromrev)) )
-         && $self->{source} ne $self->{source_root}
+# WTF do we need to check this?
+#           || !(defined $fromrev && $self->find_local_rev($fromrev, $self->{rsource_uuid}))
+         )
+         && $self->{rsource} ne $self->{rsource_root}
        ) {
 	(undef, $editor->{anchor}, $editor->{target})
-	    = File::Spec->splitpath($editor->{anchor} || $self->{source});
+	    = File::Spec::Unix->splitpath($editor->{anchor} || $self->{rsource});
 	chop $editor->{anchor};
-	$ra = $self->_new_ra ( url => $editor->{anchor}, pool => SVN::Pool->new );
+	$ra = $self->_new_ra ( url => $editor->{anchor} );
+	undef $self->{cached_ra}; # bizzare perlgc
 	@{$self}{qw/cached_ra cached_ra_url/} = ($ra, $editor->{anchor});
     }
-
+    $self->{cached_life} ||= 100;
     $editor->{target} ||= '' if $SVN::Core::VERSION gt '0.36.0';
 
 =begin NOTES
@@ -158,71 +412,103 @@ The structure of mod_lists:
 =cut
 
     $editor->{mod_lists} = {};
-    my %not_processed;
     foreach ( keys %$paths ) {
         my $item = $paths->{$_};
+	s/\n/ /g; # XXX: strange edge case
         my $href;
 
         my $svn_lpath = my $local_path = $_;
         if ( $editor->{anchor} ) {
-            $svn_lpath = $self->{source_root} . $svn_lpath;
+            $svn_lpath = $self->{rsource_root} . $svn_lpath;
             $svn_lpath =~ s|^\Q$editor->{anchor}\E/?||;
-            my $source_path = $self->{source_path} || "/";
-            $local_path =~ s|^$source_path|$self->{target_path}|;
+            my $source_path = $self->{rsource_path} || "/";
+            $local_path =~ s|^\Q$source_path\E|$self->{target_path}|;
         } else {
-            $svn_lpath =~ s|^$self->{source_path}/?||;
+            $svn_lpath =~ s|^\Q$self->{rsource_path}\E/?||;
             $local_path = "$self->{target_path}/$svn_lpath";
         }
 
+	my $local_rev = -1;
+	unless ($item->copyfrom_rev == -1) {
+	    $local_rev = $self->find_local_rev
+		($item->copyfrom_rev, $self->{rsource_uuid});
+	}
+	# XXX: the logic of the code here is a mess!
         my ($action, $rpath, $rrev, $lrev) =
             @$href{qw/action remote_path remote_rev local_rev local_path/} =
                 ( $item->action,
                   $item->copyfrom_path,
                   $item->copyfrom_rev,
-                  ($item->copyfrom_rev == -1) ?
-                    -1 : $self->find_local_rev ($item->copyfrom_rev) || undef,
+		  $local_rev,
                   $local_path,
                 );
-
+	# workaround fsfs remoet_path inconsistencies
+	$rpath = "/$rpath" if $rpath && substr ($rpath, 0, 1) ne '/';
         my ($src_lpath, $source_node_kind) = (undef, $SVN::Node::unknown);
+	# XXX: should check if the copy is within the anchor before resolving lrev
         if ( defined $lrev && $lrev != -1 ) {
-            my $rev_root = $self->{fs}->revision_root ($lrev);
-            $src_lpath = $rpath;
-            $src_lpath =~ s|^$self->{source_path}|$self->{target_path}|;
-            $source_node_kind = $rev_root->check_path ($src_lpath);
-
-            if ( $source_node_kind == $SVN::Node::none ) {
-                # The source is not in local depot.  Invalidate this
-                # copy.
-                $href->{local_rev} = undef;
-                $src_lpath = undef;
-            }
+	    $src_lpath = $rpath;
+	    # copy within mirror anchor
+            if ($src_lpath =~ s|^\Q$self->{rsource_path}\E/|$self->{target_path}/|) {
+		# $source_node_kind is used for deciding if we need reporter later
+		my $rev_root = $self->{fs}->revision_root ($lrev);
+		$source_node_kind = $rev_root->check_path ($src_lpath);
+	    }
+	    else {
+		($src_lpath, $href->{local_rev}) = (undef, undef);
+	    }
+	}
+	elsif ($rrev != -1) {
+	    # The source is not in local depot.  Invalidate this
+	    # copy.
+	    ($src_lpath, $href->{local_rev}) =
+		$self->{cb_copy_notify}
+		? $self->{cb_copy_notify}->($self, $local_path, $rpath, $rrev)
+		: (undef, undef)
         }
         @$href{qw/local_source_path source_node_kind/} =
             ( $src_lpath, $source_node_kind );
 
-        if ( /^$self->{source_path}/ ) {
+	# XXX: the loop should not reached here if changed path is
+	# not interesting to us, skip them at the beginning the the loop
+        if ( $_ eq $self->{rsource_path} or
+	     index ("$_/", "$self->{rsource_path}/") == 0 ) {
             $editor->{mod_lists}{$svn_lpath} = $href;
-        } else {
-            $not_processed{$svn_lpath} = $href;
+        } elsif ($rrev != -1 && $href->{action} eq 'A' &&
+		 index ($self->{rsource_path}, "$_/") == 0) {
+	    # special case for the parent of the anchor is copied.
+	    my $reanchor = $self->{rsource_path};
+	    $reanchor =~ s{^\Q$_\E/}{};
+	    $href->{remote_path} .= '/'.$reanchor;
+	    $href->{local_path} = $self->{target_path};
+            $editor->{mod_lists}{length $svn_lpath ? "$svn_lpath/$reanchor"
+				     : $reanchor} = $href;
         }
     }
 
     unless (keys %{$editor->{mod_lists}}) {
-        print "Skipped revision $rev.\n";
+	my $root = $editor->open_root($self->{headrev});
+	$editor->change_dir_prop ($root, svm => undef);
+	$editor->close_directory($root);
+	$editor->close_edit;
     } else {
         my @mod_list = keys %{$editor->{mod_lists}};
-        if ( grep { my $href = $editor->{mod_lists}{$_};
-                    !( ( $href->{action} eq 'A'
+        if ( ($self->{skip_to} && $self->{skip_to} <= $rev) ||
+	     grep { my $href = $editor->{mod_lists}{$_};
+                    !( ( ($href->{action} eq 'A' || $href->{action} eq 'R')
                          && defined $href->{local_rev}
                          && $href->{local_rev} != -1
                          && $href->{source_node_kind} == $SVN::Node::dir)
                        || $href->{action} eq 'D' )
                 } @mod_list ) {
+	    my $pool = SVN::Pool->new_default_sub;
+
             my $start = $fromrev || ($self->{skip_to} ? $fromrev : $rev-1);
             my $reporter =
                 $ra->do_update ($rev, $editor->{target} || '', 1, $editor);
-            $reporter->set_path ('', $start, 0);
+	    my @lock = $SVN::Core::VERSION ge '1.2.0' ? (undef) : ();
+            $reporter->set_path ('', $start, @lock, 0);
+
             $reporter->finish_report ();
         } else {
             # Copies only.  Don't bother fetching full diff through network.
@@ -237,15 +523,16 @@ The structure of mod_lists:
                 my $href = $editor->{mod_lists}{$_};
                 my $action = $href->{action};
 
-                if ($action eq 'A') {
+		if ($action eq 'D' || $action eq 'R') {
+                    $edit->delete_entry ($_);
+                }
+
+                if ($action eq 'A' || $action eq 'R') {
                     $edit->copy_directory ($_, $href->{local_source_path},
                                            $href->{local_rev});
                     $edit->close_directory ($_);
-                } elsif ($action eq 'D') {
-                    $edit->delete_entry ($_);
-                }
-            }
-
+		}
+	    }
             $edit->close_edit ();
         }
     }
@@ -260,68 +547,172 @@ The structure of mod_lists:
     }
 }
 
+sub _relayed { $_[0]->{rsource} ne $_[0]->{source} }
+
 sub get_merge_back_editor {
     my ($self, $path, $msg, $committed) = @_;
-    # get ra commit editor for $self->{source}
-    my $ra = $self->_new_ra ( url => $path ? "$self->{source}/$path" : $self->{source} );
-    my $youngest_rev = $ra->get_latest_revnum;
+    die "relayed merge back not supported yet" if $self->_relayed;
+    @{$self}{qw/cached_ra cached_ra_url/} =
+	($self->_new_ra ( url => "$self->{source}$path"), "$self->{source}$path" );
 
-    return ($youngest_rev,
-	    SVN::Delta::Editor->new ($ra->get_commit_editor ($msg, $committed)));
+    $self->{commit_ra} = $self->{cached_ra};
+    $self->load_fromrev;
+    my @lock = $SVN::Core::VERSION ge '1.2.0' ? (undef, 0) : ();
+    return ($self->{fromrev}, SVN::Delta::Editor->new
+	    ($self->{cached_ra}->get_commit_editor ($msg, $committed, @lock)));
+}
+
+sub switch {
+    my ($self, $url) = @_;
+    my $ra = $self->_new_ra (url => $url);
+    # XXX: get proper uuid like init_state
+    die "uuid is different" unless $ra->get_uuid eq $self->{source_uuid};
+    # warn "===> switching from $self->{source} to $url";
+    # get a txn, change rsource and rsource_uuidto new url
+}
+
+sub get_latest_rev {
+    my ($self, $ra) = @_;
+    # don't care about real last-modified rev num unless in skip to mode.
+    return $ra->get_latest_revnum
+	unless $self->{skip_to};
+    my ($rev, $headrev);
+    my $offset = 2;
+
+    until (defined $rev) {
+	# there were once get_log2, but it then was refactored by the svn_ra
+	# overhaul.  We have to check the version.
+	# also, it's harmful to make use of the limited get_log for svn 1.2
+	# vs svnserve 1.1, it retrieves all logs and leave the connection
+	# in an inconsistent state.
+	if ($SVN::Core::VERSION ge '1.2.0' && $self->{rsource} !~ m/^svn/) {
+	    $ra->get_log ([''], -1, 1, 1, 0, 1,
+			   sub { $rev = $_[1] });
+	}
+	else {
+	    $headrev = $ra->get_latest_revnum
+		unless defined $headrev;
+
+	    $headrev -= $offset;
+	    $ra->get_log ([''], -1, $headrev,
+			  ($SVN::Core::VERSION ge '1.2.0') ? (0) : (),
+			  0, 1,
+			  sub { $rev = $_[1] unless defined $rev});
+	    if ( $offset < $headrev ) {
+		$offset*=2;
+	    }
+	    else {
+		$offset = 2;
+	    }
+	}
+    }
+
+    die 'fatal: unable to find last-modified revision'
+	unless defined $rev;
+    return $rev;
 }
 
 sub run {
     my $self = shift;
-    my $startrev = ($self->{skip_to} || 1)-1;
+    my $ra = $self->_new_ra;
+    my $latestrev = $self->get_latest_rev ($ra);
+
+    $self->lock ('sync');
+    $self->load_fromrev;
+    # there were code here to use find_local_rev, but it will get base that
+    # is too old for use, if there are relocate happening.
+    # but this might cause race condition, while we also have lock now, need
+    # to take a closer look.
+    $self->{headrev} = $self->{fs}->youngest_rev;
+    if ($self->{skip_to} && $self->{skip_to} =~ m/^HEAD(?:-(\d+))?/) {
+	$self->{skip_to} = $latestrev - ($1 || 0);
+    }
+    my $startrev = ($self->{skip_to} || 0);
     $startrev = $self->{fromrev}+1 if $self->{fromrev}+1 > $startrev;
     my $endrev = shift || -1;
-    my $ra = $self->_new_ra;
-    $endrev = $ra->get_latest_revnum () if $endrev == -1;
+    if ($endrev && $endrev =~ m/^HEAD(?:-(\d+))?/) {
+        $endrev = $latestrev - ($1 || 0);
+    }
+    $endrev = $latestrev if $endrev == -1;
 
-    print "Syncing $self->{source}\n";
+    print "Syncing $self->{source}".($self->_relayed ? " via $self->{rsource}\n" : "\n");
 
-    return unless $endrev == -1 || $startrev <= $endrev;
+    $self->unlock ('sync'), return
+	unless $endrev == -1 || $startrev <= $endrev;
 
     print "Retrieving log information from $startrev to $endrev\n";
 
-    $self->{headrev} = $self->{fs}->youngest_rev;
-    $ra->get_log ([''], $startrev, $endrev, 1, 1,
+    my $firsttime = 1;
+    eval {
+    $ra->get_log ([''], $startrev, $endrev,
+		  ($SVN::Core::VERSION ge '1.2.0') ? (0) : (),
+		  1, 1,
 		  sub {
 		      my ($paths, $rev, $author, $date, $msg, $pool) = @_;
+		      # for the first time, skip_to might not hit
+		      # active revision in the tree. adjust to make it so.
+		      if ($firsttime) {
+			  $self->{skip_to} = $rev if defined $self->{skip_to};
+			  $firsttime = 0;
+		      }
 		      # move the anchor detection stuff to &mirror ?
-		      if (defined $self->{skip_to} && $rev < $self->{skip_to}) {
+		      if (defined $self->{skip_to} && $rev <= $self->{skip_to}) {
+			  # XXX: get the logs for skipped changes
+			  $self->{rev_incomplete} = 1;
 			  $author = 'svm';
 			  $msg = sprintf('SVM: skipping changes %d-%d for %s',
-					 $self->{fromrev}, $rev, $self->{source});
+					 $self->{fromrev}, $rev, $self->{rsource});
+		      }
+		      else {
+			  delete $self->{rev_incomplete};
 		      }
 		      $self->mirror($self->{fromrev}, $paths, $rev, $author,
 				    $date, $msg, $pool);
 		      $self->{fromrev} = $rev;
 		  });
+    };
+
+    delete $self->{cached_ra};
+    delete $self->{cached_ra_url};
+
+    $self->unlock ('sync');
+
+    return unless $@;
+    if ($@ =~ /no item/) {
+	print "Mirror source already removed.\n";
+	undef $@;
+    }
+    else {
+	die $@;
+    }
+}
+
+sub DESTROY {
 }
 
 package SVN::Mirror::Ra::MirrorEditor;
 our @ISA = ('SVN::Delta::Editor');
 use strict;
-my $debug = 0;
+our $debug = 0;
 
 =begin NOTES
 
-1. The path passed to methods in MirrorEditor is a relative directory.
-   to $self->{anchor} (if exists) or $self->{mirror}{source_root}.
+1. The path passed to methods in MirrorEditor is a relative path
+   under $self->{anchor} (if exists) or $self->{mirror}{source_root}.
 
 2. MirrorEditor can fetch usable information in $self->{mod_lists} if
    exists.
 
 3. If we need to call other method in MirrorEditor, the path must be
-   in the style of note 1.
+   in the style of #1 above.
 
 4. The diff text passed through network is patch-like.  If a directory
-   is copied, there will be one delete_entry() call for the original
+   is copied, there needs to be one delete_entry() call for the original
    directory, then a LOT of add_directory() and add_file() calls for
    each one of directories and files underneath the new directory.
+
    This behavior is easy to handle for a real file system, but hard to
-   work correct for another subversion reporitory.  A lot of codes,
+   work correct for another subversion repository.  A lot of code below,
    especially in add_directory() and add_file(), are specialised for
    handling different conditions.
 
@@ -391,7 +782,7 @@ sub change_dir_prop {
     return unless $baton;
     return if $_[0] =~ /^svm:/;
     return $self->SUPER::change_dir_prop ($baton, @_)
-	unless $_[0] =~ /^svn:(entry|wc):/;
+	unless $_[0] =~ /^svn:(?:entry|wc):/;
 }
 
 sub change_file_prop {
@@ -400,7 +791,7 @@ sub change_file_prop {
     # filter wc specified stuff
     return unless $_[0];
     return $self->SUPER::change_file_prop (@_)
-	unless $_[1] =~ /^svn:(entry|wc):/;
+	unless $_[1] =~ /^svn:(?:entry|wc):/;
 }
 
 # From source to target.  Given a path what svn lib gives, get a path
@@ -413,11 +804,11 @@ sub _translate_rel_path {
     } else {
         if ( $self->{anchor} ) {
             $path = "$self->{anchor}/$path";
-            $path =~ s|\Q$self->{mirror}{source_root}\E||;
+            $path =~ s|\Q$self->{mirror}{rsource_root}\E||;
         } else {
-            $path = "$self->{mirror}{source_path}/$path";
+            $path = "$self->{mirror}{rsource_path}/$path";
         }
-        $path =~ s|^$self->{mirror}{source_path}|$self->{mirror}{target_path}|;
+        $path =~ s|^\Q$self->{mirror}{rsource_path}\E|$self->{mirror}{target_path}|;
         return $path;
     }
 
@@ -428,7 +819,7 @@ sub _remove_entries_in_path {
 
     foreach ( sort grep $self->{mod_lists}{$_}{action} eq 'D',
               keys %{$self->{mod_lists}} ) {
-        next unless m{^$path/([^/]+)$};
+        next unless m{^\Q$path\E/([^/]+)$};
         $self->delete_entry ($_, -1, $pb, $pool);
     }
 }
@@ -599,7 +990,7 @@ sub add_directory {
     my $is_copy = 0;
     my $action = $self->_in_modified_list ($path);
     if (defined $self->{mirror}{skip_to} &&
-        $self->{mirror}{skip_to} > $self->{mirror}{working}) {
+        $self->{mirror}{skip_to} >= $self->{mirror}{working}) {
         # no-op.
     } elsif ($self->_under_latest_copypath ($path)
             && $self->_is_under_false_copy ($path)) {
@@ -609,7 +1000,7 @@ sub add_directory {
              && $self->_is_under_true_copy ($path)
              && !$action) {
         if ($self->_contains_mod_in_path ($path)) {
-            splice @_, 0, 2, $self->{mirror}{headrev};
+            $is_copy = 1;
             $method = 'open_directory';
             $self->_enter_new_copied_path ();
         } else {
@@ -638,7 +1029,6 @@ sub add_directory {
             }
         } elsif ($action eq 'M') {
             # 1d.
-            splice @_, 0, 2, $self->{mirror}{headrev};
             $method = 'open_directory';
 
             $self->_enter_new_copied_path ()
@@ -657,17 +1047,17 @@ sub add_directory {
             } elsif ( !defined ($copyrev) ) {
                 $self->_create_new_copied_path ($path);
             }
+	    else {
+		$self->_enter_new_copied_path ()
+		    if $self->_under_latest_copypath ($path);
+	    }
         }
     } else {
         # raise error.
         die "Oh no, no more exceptions!  add_directory() failed.";
     }
 
-    if ($path eq $self->{target}) {
-        splice @_, 0, 2, $self->{mirror}{headrev}
-            if $method ne "open_directory";
-        $method = "open_directory";
-    }
+    $method = "open_directory" if $path eq $self->{target};
 
     push @{$self->{visited_paths}}, $path;
     my $tran_path = $self->_translate_rel_path ($path);
@@ -675,10 +1065,16 @@ sub add_directory {
     $method = 'open_directory'
         if $tran_path eq $self->{mirror}{target_path};
 
+    splice @_, 0, 2, $self->{mirror}{headrev}
+	if $method eq "open_directory";
+
     my $dir_baton;
     $method = "SUPER::$method";
     $dir_baton = $self->$method($tran_path, $pb, @_);
 
+    # Always 'touch' the directory, even for empty modifications.
+    $self->change_dir_prop ( $dir_baton, 'svm' => undef, $pool );
+
     $self->_remove_entries_in_path ($path, $dir_baton, $pool) if $is_copy;
 
     return $dir_baton;
@@ -703,7 +1099,7 @@ sub close_directory {
     $self->_leave_new_copied_path ()
         if $self->_under_latest_copypath ($path);
         
-    $self->SUPER::close_directory ($baton);
+    $self->SUPER::close_directory ($baton, @_);
 }
 
 sub close_file {
@@ -745,7 +1141,7 @@ sub add_file {
     my $method = 'add_file';
     my $action = $self->_in_modified_list ($path);
     if ((defined $self->{mirror}{skip_to}
-         && $self->{mirror}{skip_to} > $self->{mirror}{working})
+         && $self->{mirror}{skip_to} >= $self->{mirror}{working})
         || ($self->_under_latest_copypath ($path)
             && $self->_is_under_false_copy ($path)) ) {
         # no-op. add_file().
@@ -792,30 +1188,26 @@ sub delete_entry {
     print "MirrorEditor::delete_entry($path, $rev)\n" if $debug;
     return unless $pb;
 
-    my $source_path = $self->{mirror}{source_path};
-    my $target_path = $self->{mirror}{target_path};
-    $source_path =~ s{^/}{};
-    if ($source_path && $path =~ m/^$source_path/) {
-        $path =~ s/$source_path/$target_path/;
-    } else {
-        $path = "$target_path/$path";
-    }
-
-    $self->SUPER::delete_entry ($path, $rev, $pb, $pool);
+    $self->SUPER::delete_entry ($path, $self->{mirror}{headrev},
+				$pb, $pool);
 }
 
 sub close_edit {
-    my $self = shift;
+    my ($self, $pool) = @_;
     print "MirrorEditor::close_edit()\n" if $debug;
 
     unless ($self->{root}) {
         # If we goes here, this must be an empty revision.  We must
         # replicate an empty revision as well.
-        $self->open_root ($self->{mirror}{headrev});
+        $self->open_root ($self->{mirror}{headrev}, $pool);
+	$self->SUPER::close_directory ($self->{root}, $pool);
     }
+    delete $self->{root};
+    local $SIG{INT} = 'IGNORE';
+    local $SIG{TERM} = 'IGNORE';
 
-    $self->SUPER::close_directory ($self->{root});
-    $self->SUPER::close_edit (@_);
+    $self->{mirror}->lock ('mirror');
+    $self->SUPER::close_edit ($pool);
 }
 
 1;
